SystemOrganization addCategory: #'UpdateStreamer-Core'!

CommandLineHandler subclass: #LoadUpdatesCommandLineHandler
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'UpdateStreamer-Core'!

!LoadUpdatesCommandLineHandler commentStamp: '<historical>' prior: 0!
Usage: update

Documentation:
The update the image to the latest version.
!

!LoadUpdatesCommandLineHandler class methodsFor: 'accessing' stamp: 'MarcusDenker 11/7/2012 13:57'!
commandName
	^ 'update'! !

!LoadUpdatesCommandLineHandler class methodsFor: 'accessing' stamp: 'MarcusDenker 11/7/2012 13:57'!
description
	^ 'load updates'! !

!LoadUpdatesCommandLineHandler methodsFor: 'activation' stamp: 'CamilloBruni 1/14/2013 16:12'!
activate
	self activateHelp
		ifTrue: [  ^ self ].
		
	(self hasOption: 'from-file')
		ifTrue: [ self loadUpdatesFromFile ]
		ifFalse: [ self loadUpdates ].

	Smalltalk snapshot: true andQuit: true.! !

!LoadUpdatesCommandLineHandler methodsFor: 'actions' stamp: 'MarcusDenker 11/7/2012 13:58'!
loadUpdates
	UpdateStreamer new updateFromServer! !

!LoadUpdatesCommandLineHandler methodsFor: 'actions' stamp: 'CamilloBruni 1/14/2013 16:12'!
loadUpdatesFromFile
	^ self loadUpdatesFromFile: (self optionAt: 'from-file') asFileReference ! !

!LoadUpdatesCommandLineHandler methodsFor: 'actions' stamp: 'CamilloBruni 1/14/2013 16:56'!
loadUpdatesFromFile: updatesFile
	updatesFile exists
		ifFalse: [ ^ self exitFailure: (updatesFile fullName, ' does not exist!!') ].
	
	UpdateStreamer new updateFromFile: updatesFile! !

Object subclass: #UpdateStreamer
	instanceVariableNames: 'applyChanges silent upToNumber saveLocally path pharoServer'
	classVariableNames: 'UpdateDownloader'
	poolDictionaries: ''
	category: 'UpdateStreamer-Core'!

!UpdateStreamer commentStamp: 'StephaneDucasse 9/1/2010 21:59' prior: 0!
I'm a little class inspired heavily from Utilities that download update.list and launch update.

you can tell me to update throw a given number of update or to be verbose.


UpdateStreamer new 
	beVerbose;
	updateFromServer
	
This will update all the updates available and show up dialogs
	!

!UpdateStreamer methodsFor: 'accessing' stamp: 'StephaneDucasse 9/1/2010 16:43'!
beSilent

	silent := true! !

!UpdateStreamer methodsFor: 'accessing' stamp: 'StephaneDucasse 9/1/2010 16:43'!
beVerbose

	silent := false! !

!UpdateStreamer methodsFor: 'private actions' stamp: 'CamilloBruni 1/14/2013 16:50'!
elementaryReadServerUpdates
	"Scan the update server(s) for unassimilated updates. If the instance variable upToNumber is not nil, it represents the highest-numbered update to load.  This makes it possible to update only up to a particular point."

	"UpdateStreamer new elementaryReadServerUpdates"

	^ self elementaryReadServerUpdates: self newUpdates.! !

!UpdateStreamer methodsFor: 'private actions' stamp: 'CamilloBruni 1/14/2013 16:49'!
elementaryReadServerUpdates: urls
	"Scan the update server(s) for unassimilated updates. If the instance variable upToNumber is not nil, it represents the highest-numbered update to load.  This makes it possible to update only up to a particular point."

	"UpdateStreamer new elementaryReadServerUpdates"

	| str failed loaded this nextDoc updateName |
	(Smalltalk globals includesKey: #ScriptLoader) ifFalse: [
		^ self feedback: 'ScriptLoader is not present (this may happen if you have run cleanUpForProduction). ' ].
	Cursor wait showWhile: 
		[ | docQueue docQueueSema |
		loaded := 0.
		failed := nil.
		"send downloaded documents throuh this queue"
		docQueue := SharedQueue new.
		"this semaphore keeps too many documents from beeing queueed up at a time"
		docQueueSema := Semaphore new.
		5 timesRepeat: [ docQueueSema signal ].

		"fork a process to download the updates"
		self retrieveUrls: urls ontoQueue: docQueue withWaitSema: docQueueSema.

		"process downloaded updates in the foreground"

		[ this := docQueue next.
		nextDoc := docQueue next.  
		nextDoc = #failed ifTrue: [ failed := this ].
		(failed isNil and: [ nextDoc ~= #finished ])
		] whileTrue: [
			failed ifNil: [
				nextDoc size = 0 ifTrue: [ failed := this ]. ].
			failed ifNil: [
				nextDoc peek asciiValue = 4	"pure object file"
					ifTrue: [failed := this]].	"Must be fileIn, not pure object file"
			failed ifNil: [
				self shouldApplyChanges
					ifTrue: [
						updateName := (this findTokens: '/') last.
						ChangeSet newChangesFromStream: nextDoc named: updateName.
						SystemVersion current registerUpdate: updateName initialIntegerOrNil].
			self saveLocally ifTrue:
				[self saveUpdate: nextDoc onFile: (this findTokens: '/') last].
			].
		loaded := loaded + 1.
		docQueueSema signal].
	].

	failed ~~ nil & (urls size - loaded > 0) ifTrue: [
		str := loaded printString ,' new update file(s) processed.'.
		str := str, '\Could not load ' withCRs, 
			(urls size - loaded) printString ,' update file(s).',
			'\Starting with "' withCRs, failed, '".'.
		self feedback: str].
	^ Array with: failed with: loaded
! !

!UpdateStreamer methodsFor: 'private actions' stamp: ''!
extractVersion: aVersion from: updateList
	"Pull out the part of the list that applies to a version."
	"self new extractVersion: 'Pharo1.1' from: ..."
	
	| listContents version versIndex |
	self flag: #shouldUseADictionary.
	listContents := self groupUpdatesByVersion: updateList.
	versIndex := (listContents collect: [ :pair | pair first]) indexOf: aVersion.
	versIndex = 0 ifTrue: [^ Array new].
	^ (listContents at: versIndex) last! !

!UpdateStreamer methodsFor: 'private actions' stamp: 'StephaneDucasse 9/1/2010 16:46'!
feedback: aString

	self silent 
		ifFalse: [self inform: aString]
		ifTrue: [Transcript show: aString ; cr]! !

!UpdateStreamer methodsFor: 'private actions' stamp: ''!
groupUpdatesByVersion: updateContents
	"Parse the contents of updates.list into {{vers. {fileNames*}}*}, and return it."
	
	| sections vers strm line fileNames |
	self flag: #shouldUseADictionary.
	sections := OrderedCollection new.
	fileNames := OrderedCollection new: 1000.
	strm := updateContents readStream.
	[ strm atEnd ] whileFalse: 
		[ line := strm nextLine.
		line size > 0 ifTrue: 
			[ line first = $# 
				ifTrue: 
					[ vers ifNotNil: [ sections addLast: {  vers. (fileNames asArray)  } ].
					"Start a new section"
					vers := line allButFirst.
					fileNames resetTo: 1 ]
				ifFalse: [ line first = $* ifFalse: [ fileNames addLast: line ] ] ] ].
	vers ifNotNil: [ sections addLast: {  vers. (fileNames asArray)  } ].
	^ sections 
	! !

!UpdateStreamer methodsFor: 'private actions' stamp: 'CamilloBruni 1/14/2013 16:47'!
newUpdates
	"Return a list of fully formed URLs of update files we do not yet have up to the upToNumber instance variable if set." 

	"self new newUpdates"

	| result |
	
	result := ZnClient new
		systemPolicy;
		setAcceptEncodingGzip;
		beOneShot;
		accept: ZnMimeType text;
		url: self updateListFileURLString;
		ifFail: [ :exception | self error: 'Cannot access update server: ' , exception printString ];
		get.
		
	^ self newUpdatesFromString: result.! !

!UpdateStreamer methodsFor: 'private actions' stamp: 'CamilloBruni 1/14/2013 16:47'!
newUpdatesFromFile: aFileReference
	"
	A file should be of following format:

#Pharo2.0a
20000-Pha-Empty.cs
20001-Pha-Empty2.cs
20002-Pha-DelScriptLoader.cs
	
	 "
	^self newUpdatesFromString: aFileReference readStream contents
! !

!UpdateStreamer methodsFor: 'private actions' stamp: 'CamilloBruni 1/14/2013 16:47'!
newUpdatesFromString: aString
	"
	A string should be of following format:

#Pharo2.0a
20000-Pha-Empty.cs
20001-Pha-Empty2.cs
20002-Pha-DelScriptLoader.cs
	
	 "
	
	| list out existing |
	
	out := OrderedCollection new.
	existing := SystemVersion current updates.
	list := self extractVersion: SystemVersion current version from: aString.
	
	list reverseDo: [:fileName | | itsNumber |
		itsNumber := fileName initialIntegerOrNil. 
		(existing includes: itsNumber)
			ifFalse: [ (itsNumber isNil or: [itsNumber <= self upToNumber])
							ifTrue: [out addFirst: self updateFolderURL, fileName]]
			ifTrue: [^ out]].
	"("
	(out size > 0) "or: 
			[char := doc reset; skipSeparators; next.
			(char == $*) | (char == $#)])"  ifTrue:
				[^ out "we have our list"].	
	self feedback: 'All code update servers seem to be unavailable'.
	^ out
	
! !

!UpdateStreamer methodsFor: 'private actions' stamp: 'StephaneDucasse 9/1/2010 17:32'!
newUpdatesUpToNumber: aNumber
	"Return a list of fully formed URLs of update files we do not yet have.  Look at the file 'updates.list'.
	No updates numbered higher than aNumber (if it is not nil) are returned " 

	"self new newUpdatesToNumber: 14000"

	self upToNumber: aNumber.
	^ self newUpdates.! !

!UpdateStreamer methodsFor: 'accessing' stamp: ''!
path
	"Part ... in the following  http:// server / path /updates.list"
	
	^ path ifNil: [ path := 'updates/pharo'].! !

!UpdateStreamer methodsFor: 'accessing' stamp: ''!
path: aString
	"http:// server / aString /updates.list"
	
	^ path ifNil: [ path := 'updates/pharo'].! !

!UpdateStreamer methodsFor: 'private actions' stamp: 'StephaneDucasse 9/1/2010 19:15'!
retrieveUrls: urls ontoQueue: queue withWaitSema: waitSema 
	"download the given list of URLs. The queue will be loaded alternately  
	with url's and with the retrieved contents. If a download fails, the  
	contents will be #failed. If all goes well, a special pair with an empty  
	URL and the contents #finished will be put on the queue. waitSema is  
	waited on every time before a new document is downloaded; this keeps 
	the downloader from getting too far  ahead of the main process"
	"kill the existing downloader if there is one"
	
	self silent 
		ifTrue: [self silentRetrieveUrls: urls ontoQueue: queue withWaitSema: waitSema ]
		ifFalse: [self verboseRetrieveUrls: urls ontoQueue: queue withWaitSema: waitSema ]
! !

!UpdateStreamer methodsFor: 'accessing' stamp: ''!
saveLocally

	^ saveLocally ifNil: [saveLocally := false]! !

!UpdateStreamer methodsFor: 'accessing' stamp: ''!
saveLocally: aBoolean

	saveLocally := aBoolean! !

!UpdateStreamer methodsFor: 'untested' stamp: 'SeanDeNigris 7/12/2012 08:45'!
saveUpdate: doc onFile: fileName
	"Save the update on a local file.  With or without the update number on the front"

	| file fName pos updateDirectory |
	updateDirectory := 'updates' asFileReference.
	updateDirectory ensureDirectory.

	fName := fileName.
	(self updateRemoveSequenceNum) ifTrue:
		[pos := fName findFirst: [:c | c isDigit not].
		fName := fName copyFrom: pos to: fName size].
	
	doc reset; ascii.
	
	file := updateDirectory / fName.
	file exists ifTrue: [ ^ self ].
	
	file writeStreamDo: [ :s|
		s nextPutAll: doc contents].
! !

!UpdateStreamer methodsFor: 'accessing' stamp: ''!
server

	pharoServer ifNil: [pharoServer := 'pharo.gforge.inria.fr/'].
	^ pharoServer! !

!UpdateStreamer methodsFor: 'accessing' stamp: ''!
server: aStringWithHTTPColon

	pharoServer := aStringWithHTTPColon
! !

!UpdateStreamer methodsFor: 'accessing' stamp: 'StephaneDucasse 9/1/2010 16:49'!
shouldApplyChanges
	"when true the changes are downloaded and executed in the image"
	
	^ applyChanges ifNil: [applyChanges := true ]! !

!UpdateStreamer methodsFor: 'accessing' stamp: 'StephaneDucasse 9/1/2010 16:50'!
shouldApplyChanges: aBoolean
	"when true the changes are downloaded and executed in the image"
	
	applyChanges := aBoolean ! !

!UpdateStreamer methodsFor: 'accessing' stamp: 'MarcusDenker 11/29/2010 21:54'!
silent
	
	^ silent ifNil: [self beVerbose.false.]! !

!UpdateStreamer methodsFor: 'private actions' stamp: 'CamilloBruni 10/1/2012 12:53'!
silentRetrieveUrls: urls ontoQueue: queue withWaitSema: waitSema 
	"download the given list of URLs. The queue will be loaded alternately  
	with url's and with the retrieved contents. If a download fails, the  
	contents will be #failed. If all goes well, a special pair with an empty  
	URL and the contents #finished will be put on the queue. waitSema is  
	waited on every time before a new document is downloaded; this keeps 
	the downloader from getting too far  ahead of the main process"
	"kill the existing downloader if there is one"
	
	self zapUpdateDownloader.
	"fork a new downloading process"
	UpdateDownloader := [
		urls do: [ :url | | doc | 
			waitSema wait.
			queue nextPut: url.
			doc := ZnClient new
				systemPolicy;
				setAcceptEncodingGzip;
				beOneShot;
				ifFail: [ :exception | 
					queue nextPut: #failed.
					UpdateDownloader := nil.
					Processor activeProcess terminate ];
				get: url.
			queue nextPut: doc readStream ].		
		queue nextPut: ''.
		queue nextPut: #finished.
		UpdateDownloader := nil] newProcess.
	UpdateDownloader priority: Processor userInterruptPriority.
	"start the process running"
	UpdateDownloader resume! !

!UpdateStreamer methodsFor: 'accessing' stamp: 'StephaneDucasse 9/1/2010 17:30'!
upToNumber
	"when specified the update will stop at this number"
	
	^ upToNumber ifNil: [ upToNumber := 1000000].! !

!UpdateStreamer methodsFor: 'accessing' stamp: 'StephaneDucasse 9/1/2010 16:57'!
upToNumber: aNumber
	"when specified the update will stop at this number"
	
	upToNumber := aNumber. ! !

!UpdateStreamer methodsFor: 'accessing' stamp: 'StephaneDucasse 1/31/2011 22:26'!
updateFolderURL
	"in the future it may be wiser to use SystemVersion to avoid to duplicate the logic"
	"SystemVersion current major"
	^ 'http://', self server, self path, (Smalltalk globals at: #ScriptLoader) currentMajorVersionNumber asString, '/'! !

!UpdateStreamer methodsFor: 'public' stamp: 'CamilloBruni 1/14/2013 16:55'!
updateFromFile: aFileReference
	self updateWith: [ self elementaryReadServerUpdates: (self newUpdatesFromFile: aFileReference) ]! !

!UpdateStreamer methodsFor: 'public' stamp: 'CamilloBruni 1/14/2013 16:55'!
updateFromServer
	self updateWith: [ self elementaryReadServerUpdates ]! !

!UpdateStreamer methodsFor: 'updating' stamp: 'Integrator 6/21/2012 16:27'!
updateFromServerSilently
	"Scan the update server for unassimilated updates. If upToNumber is not nil, it represents the highest-numbered update to load.  This makes it possible to update only up to a particular point.   If saveLocally is true, then save local copies of the update files on disc.  If shouldApplyChanges is true, then absorb the updates into the current image.

A file on the server called updates.list has the names of the last N update files.  We look backwards for the first one we do not have, and start there"
"* To add a new update:  Name it starting with a new two-digit code.  
* Do not use %, /, *, space, or more than one period in the name of an update file.
* The update name does not need to have any relation to the version name.
* Figure out which versions of the system the update makes sense for.
* Add the name of the file to each version's category below.
* Put this file and the update file on all of the servers.
*
* To make a new version of the system:  Pick a name for it (no restrictions)
* Put # and exactly that name on a new line at the end of this file.
* During the release process, fill in exactly that name in the dialog box.
* Put this file on the server."

"When two sets of updates need to use the same directory, one of them has a * in its 
serverUrls description.  When that is true, the first word of the description is put on
the front of 'updates.list', and that is the index file used."

	| pair failed loaded |
	pair := self elementaryReadServerUpdates.
	failed := pair first.
	loaded := pair second.
	^ true! !

!UpdateStreamer methodsFor: 'public' stamp: 'CamilloBruni 1/14/2013 16:56'!
updateFromString: aString
	self updateWith: [ self elementaryReadServerUpdates: (self newUpdatesFromString: aString) ]! !

!UpdateStreamer methodsFor: 'accessing' stamp: 'StephaneDucasse 1/31/2011 22:24'!
updateListFileURLString
	"in the future it may be wiser to use VersionNumber to avoid to duplicate the logic"
	
	^ self updateFolderURL, self updatesListFileName! !

!UpdateStreamer methodsFor: 'public' stamp: 'CamilloBruni 1/14/2013 16:55'!
updateWith: anUpdateBlock

	"Scan the update server for unassimilated updates. If upToNumber is not nil, it represents the highest-numbered update to load.  This makes it possible to update only up to a particular point.   If saveLocally is true, then save local copies of the update files on disc.  If shouldApplyChanges is true, then absorb the updates into the current image.

A file on the server called updates.list has the names of the last N update files.  We look backwards for the first one we do not have, and start there"
"* To add a new update:  Name it starting with a new two-digit code.  
* Do not use %, /, *, space, or more than one period in the name of an update file.
* The update name does not need to have any relation to the version name.
* Figure out which versions of the system the update makes sense for.
* Add the name of the file to each version's category below.
* Put this file and the update file on all of the servers.
*
* To make a new version of the system:  Pick a name for it (no restrictions)
* Put # and exactly that name on a new line at the end of this file.
* During the release process, fill in exactly that name in the dialog box.
* Put this file on the server."

"When two sets of updates need to use the same directory, one of them has a * in its 
serverUrls description.  When that is true, the first word of the description is put on
the front of 'updates.list', and that is the index file used."
	[[[
		| pair failed loaded |
		pair := anUpdateBlock value.
		failed := pair first.
		loaded := pair second.
		failed ifNil: ["is OK"
			 self feedback:  (loaded printString ,' new update file(s) processed.')]] 
	on: (Smalltalk at: #MCMergeOrLoadWarning ifAbsent: [ nil ]) do: [ :e| e resume: false ]] 
	on: (Smalltalk at: #MCNoChangesException ifAbsent: [ nil ]) do: [ :e| e resume]]
 	on: MCMergeResolutionRequest do: [ :request |
			request merger conflicts isEmpty
				ifTrue: [ request resume: true ]
				ifFalse: [ request pass ]].! !

!UpdateStreamer methodsFor: 'accessing' stamp: 'SeanDeNigris 7/12/2012 08:44'!
updatesListFile
	^  self updatesListFileName asFileReference! !

!UpdateStreamer methodsFor: 'accessing' stamp: 'StephaneDucasse 2/10/2011 15:39'!
updatesListFileName

	^ 'updates', (SystemVersion current majorMinor:''), '.list'! !

!UpdateStreamer methodsFor: 'private actions' stamp: 'CamilloBruni 10/1/2012 12:53'!
verboseRetrieveUrls: urls ontoQueue: queue withWaitSema: waitSema 
	"download the given list of URLs. The queue will be loaded alternately  
	with url's and with the retrieved contents. If a download fails, the  
	contents will be #failed. If all goes well, a special pair with an empty  
	URL and the contents #finished will be put on the queue. waitSema is  
	waited on every time before a new document is downloaded; this keeps 
	the downloader from getting too far  ahead of the main process"
	"kill the existing downloader if there is one"
	
	self zapUpdateDownloader.
	"fork a new downloading process"
	UpdateDownloader := [ | updateCounter |
		updateCounter := 0.
		'Downloading updates' displayProgressFrom: 0 to: urls size during: [:bar |
			urls do: [ :url | | doc | 
				waitSema wait.
				queue nextPut: url.
				doc := ZnClient new
					systemPolicy;
					setAcceptEncodingGzip;
					beOneShot;
					ifFail: [ :exception | 
						queue nextPut: #failed.
						UpdateDownloader := nil.
						Processor activeProcess terminate ];
					get: url.
				queue nextPut: doc readStream. 
				updateCounter := updateCounter + 1. 
				bar current: updateCounter ] ].
			queue nextPut: ''.
			queue nextPut: #finished.
			UpdateDownloader := nil ] newProcess.
	UpdateDownloader priority: Processor userInterruptPriority.
	"start the process running"
	UpdateDownloader resume ! !

!UpdateStreamer methodsFor: 'untested' stamp: ''!
writeList: listContents toStream: strm
	"Write a parsed updates.list out as text.
	This is the inverse of groupUpdatesByVersion:"

	"so far did not tested it - sd"
	[strm reset.
	listContents do:
		[:pair | | fileNames version | 
				version := pair first.  
				fileNames := pair last.
		strm nextPut: $#; nextPutAll: version; cr.
		fileNames do: [:fileName | strm nextPutAll: fileName; cr]]]
	ensure: [strm close]! !

!UpdateStreamer methodsFor: 'private actions' stamp: ''!
zapUpdateDownloader

	UpdateDownloader ifNotNil: [UpdateDownloader terminate].
	UpdateDownloader := nil.! !
